"""
This module provides the main class for the kinematic GPS processing using
the TRACK program from the GAMIT/GLOBK suite.

"""

__author__    = "Patrick Riesen (riesen@vaw.baug.ethz.ch)"
__date__      = "2010-07-01"
__copyright__ = "Copyright (C) 2010 Patrick Riesen"
__license__   = "GNU GPL version 3 or any later version"

import sys
import os, getopt
import commands as cm
import numpy as n
import string as st
import pickle as pk
from datetime import datetime as dt

from kingps.rinexconverter import toYearFraction
from kingps.variables import execute
#-------------------------------------------------------------------------------------
class GPSKinematicProcessing:
    """
    Class for GPS Kinematic Processing with Track from GAMIT/Globk

    This class must be instantiated with a GPSProcessingVariables object.
    """
#   ----------------------------------------           
    def __init__(self, gpsvariables):
        """ Initialize local variables for this class. """
        
        self.main = gpsvariables
        self.reference_station_files = []
        self.processing_files = []
        self.processing_files_date = []
        self.processing_files_dirs = []
        self.processing_files_orbs = []
        self.files_unable_to_process = []

#   ----------------------------------------           
    def get_files_to_process(self, site):
        """
        Fetch the files for processing for each (daily) file
        and make sure data is prossible to process by checking if
        data from the reference station is also available +/- 1day.

        Args:
            site   = A station name (string) which belongs to
                     the list of the sites.                   
        """

        # fetch processing files
        os.chdir(self.main.rinex_output_dir + "/" + site)
        print self.main.rinex_output_dir + "/" + site
        all_files = []
        for fn in os.listdir("."):
            if fn.startswith(site) and fn.endswith("o"):
                all_files += [fn,]
                
        # filestring = cm.getoutput("ls -l *.*o | awk '{print $8}'")
        # all_files = st.split(filestring, sep='\n')        

        self.reference_station_files = []
        self.processing_files = []
        self.processing_files_date = []
        self.files_unable_to_process = []

        for fn in all_files:
            # get year and day from filename
            [doy, year] = fn[len(site):].rstrip("o").split(".")
            pfn = self.main.rinex_output_dir + "/" + site + "/" + fn

            # check reference station files which should be there on the same day
            rfn = self.main.fix_name + st.zfill(str(doy),3) + "." + year + "o"
            rdirfn = self.main.rinex_output_dir + "/" + self.main.fix_name + "/" + rfn

            if(os.path.isfile(rdirfn)):

                # fetch decimal year date
                grep_date = "sed -n '/L1 & L2/{n;p;}' " + pfn \
                            + " | awk '{print $1,$2,$3}'"

                datestr = cm.getoutput(grep_date)
                dateobj = dt.strptime(datestr,"%Y %m %d")                

                self.processing_files_date.append(float(toYearFraction(dateobj)))
                
                # get_year = cm.getoutput("doy " + cm.getoutput(grep_date) \
                #                         + " | grep 'Decimal' | awk '{print $3}'")
                self.processing_files.append(fn)
                # self.processing_files_date.append(float(get_year))
                self.reference_station_files.append(rfn)
            else:
                self.files_unable_to_process.append(fn)
                print "\n*** Warning: The file",fn,"can not be processed."
                print "    No reference station data!"

        # if some files could not be processed, write the list
        # to a file
        if(len(self.files_unable_to_process) > 0):
            os.chdir(self.main.processing_output_dir)
            misslogfile = site + "_unprocessed.log"
            f = open(misslogfile,"w")
            for mfn in self.files_unable_to_process:
                f.write(mfn + "\n")
            f.close()

            print "\n*** Warning: Some files could not be processed. Please have a"
            print "    look at",misslogfile,"."

        os.chdir(self.main.home_dir)
            
#   ----------------------------------------           
    def fetch_orbit_files(self, year, doy, ndays=3):
        """
        Fetch orbit files for relevant data.

        Args:
            year   = string of the year
            doy    = string of day of year
            ndays  = integer of the number of the range
                     of days around the day specified
                     orbit files should be fetched.
                     e.g., if ndays = 3 and doy is 63,
                     orbit files for doy 62, 63, 64
                     will be fetched from the FTP repositories.
                     (defaults to 3)
        """

        pwd = os.getcwd()
        os.chdir(self.main.orbit_files_dir)
        self.processing_files_orbs = []

        # fetch files for +/-ndays of processing day
        arnd = n.arange(0 - (ndays-2) , ndays-1).tolist()
        for ia in arnd:
            fetch_doy = int(doy) + ia
            fetch_year = int(year)

            # first or last day of year special issue
            if(doy == 1):
                fetch_year = int(year) - 1
                fetch_doy = 365
            elif(doy == 365):
                fetch_year = int(year) + 1
                fetch_doy = 1

            # command for fetching
            orbcmd = self.main.orbits + " -archive " + self.main.orbits_archive + \
                     " -yr " + str(fetch_year) + " -doy " + str(int(fetch_doy)) + \
                     " -pre f -makeg no -type " + self.main.orbits_type[1]

            # check if file exists already,
            # only download it when it's not already there
            gpsweek = cm.getoutput("doy " + str(fetch_year) + " " + str(fetch_doy) \
                                   + " | grep 'GPS Week' | awk '{print $3$7}'").rstrip(",")
            orbitfile = self.main.orbits_type[0] + gpsweek + "." + self.main.orbits_type[1]
            self.processing_files_orbs.append(orbitfile)
            
            if(os.path.isfile(orbitfile) == False):
                self.main.thread(execute(orbcmd))

        # Clean away log files from FTP
        cleancmd = "rm -Rf *.log"
        self.main.thread(execute(cleancmd))
        os.chdir(pwd)


#   ----------------------------------------           
    def init_daily_workspaces(self, site):
        """
        Create subfolders of days for the specified site and prepare
        for processing

        Args:
            site   = A station name (string) which belongs to
                     the list of the sites.                   

        """

        print "\n-----------------------------------------------------------"
        print "Prepare data for processing for site: ",site

        # Check if processing directory has been set.
        ppdir = self.main.processing_output_dir + "/" + site
        try:
            os.chdir(ppdir)
        except OSError:
            if(site == self.main.fix_name):
                print "*** Error: The reference station does not need to be processed."
                raise os.error()
            else:
                print "*** Error: Could not create or access output processing directory",ppdir
                raise os.error()

        self.processing_files_dirs = []

        # Create folders site/year/day for processing
        for date, rfile, pfile in zip(self.processing_files_date,
                                      self.reference_station_files,
                                      self.processing_files):

            [doy, year] = pfile[len(site):].rstrip("o").split(".")
            # make 4-digit year
            # Y = cm.getoutput("doy " + year + " " + doy + " | grep Decimal | awk '{print $3}'")
            year = int(n.fix(float(date)))
            pfdir = self.main.processing_output_dir + "/" + site + "/" + str(year) + "/" + doy

            # copy gps rinex data in there
            if(os.path.isdir(pfdir) == False):
                os.makedirs(pfdir)
            cpyfns = [self.main.rinex_output_dir + "/" + self.main.fix_name + "/" + rfile,
                      self.main.rinex_output_dir + "/" + site + "/" + pfile]
            for cf in cpyfns:
                cpycmd = "cp " + cf + " " + pfdir
                execute(cpycmd)

            # copy the correct orbit files in there
            self.fetch_orbit_files(year, doy)
            for ofn in self.processing_files_orbs:
                get_orbfile_cmd = "cp " + self.main.orbit_files_dir + "/" + ofn \
                              + " " + pfdir
                execute(get_orbfile_cmd)

            pwd = os.getcwd()
            os.chdir(pfdir)
            os.system("cat *." + self.main.orbits_type[1] + " > orbitfile.dat")
            #os.system("rm -Rf " + self.main.orbits_type[0] + "*")
            os.chdir(pwd)

            # remember the folders to process
            self.processing_files_dirs.append(pfdir)

        os.chdir(self.main.home_dir)

#   ----------------------------------------           
    def create_batch_files(self, site):
        """
        Create command files for daily batch processing with Track.

        Args:
            site   = A station name (string) which belongs to
                     the list of the sites.                   
        """

        # These lines go into the batch file anyway
        rname = self.main.fix_name
        l0 = " nav_file orbitfile.dat " + self.main.orbits_type[1]
        l1 = " out_type " + self.main.output
        l2 = " mode " + self.main.mode
        batchfile = [l0, l1, l2]

        # Specify position coordinates for reference station
        if(self.main.fix_xyz == None):
            print "\n*** Warning: No coordinates specified for reference station, using XYZ coordinates"
            print "    from the Rinex file header."
        else:
            l3 = " site_pos\n   " + self.main.fix_name
            plch = " x y z"
            coord = ["x", "y", "z"]
            for c, fc in zip(coord, self.main.fix_xyz):
                plch = plch.replace(c, str(fc))
            batchfile.append(l3+plch)

        # fill in additional track commands
        if(self.main.site_stats != None):
            l4 = " site_stats\n   " + site \
                 + " s1 s1 s1  s2 s2 s2".replace("s1", \
                                                 str(self.main.site_stats[0])).replace("s2", \
                                                                                       str(self.main.site_stats[1]))
            batchfile.append(l4)

        if(self.main.smooth):
            l5 = " back_type smooth"
            batchfile.append(l5)

        if(self.main.interval != None):
            l6 = " interval " + str(self.main.interval)
            batchfile.append(l6)

        if(self.main.out_sig_limit != None):
            l7 = " out_sig_limit " + str(self.main.out_sig_limit)
            batchfile.append(l7)

        if(self.main.cut_off != None):
            l8 = " cut_off " + str(self.main.cut_off)
            batchfile.append(l8)

        # write the batch files into processing directories
        for rf, pf, pd in zip(self.reference_station_files,
                         self.processing_files,
                         self.processing_files_dirs):
            R = " obs_file\n   " + rname + " " + rf + " F"
            K = "   " + site + " " + pf + " K"
            batchfile_header = R + "\n" + K

            fn = open(pd + "/track.bat", 'w')
            fn.write(batchfile_header + "\n\n")

            for line in batchfile:
                fn.write(line + "\n\n")
            fn.close()

#   ----------------------------------------           
    def estimate_positions(self, site):
        """
        Go through folder structure and run Track in each of the daily folders.

        Args:
            site   = A station name (string) which belongs to
                     the list of the sites.                   
        """

        print "\n--------- Track GPS Processing ----------------------------\n"

        # Execute track in each processing directory
        for pf in self.processing_files_dirs:
            os.chdir(pf)
            run_cmd = self.main.track + " -f track.bat >&track.log"
            print "working on folder",pf
            self.main.thread(execute(run_cmd))

#   ----------------------------------------           
    def process(self, which="all", cleanup=False):
        """
        Process the available data for all or a single site specified.

        Args:
            which = A string which can be either 'all' (default) or
                a name of a site specified for processing.
                cleanup = if True, all temporary files (e.g., copies from the
                rinex files, track logging/sumary output files etc.)
                will be removed. Defaults to False.
        """
        # run relevant functions for all sites specified
        if(which == "all"):
            if(self.main.site_names != None):
                for site in self.main.site_names:
                    self.get_files_to_process(site)
                    self.init_daily_workspaces(site)
                    self.create_batch_files(site)
                    self.estimate_positions(site)
                    if(cleanup):
                        cleanup(site)
        else:
            # try to execute commans for the given site, only if it
            # is defined in site_nams
            try:
                self.main.site_names.index(which)
            except ValueError:
                print "*** Error: This site was not specified (site_names)."
                raise ValueError
            else:
                site = which
                self.get_files_to_process(site)
                self.init_daily_workspaces(site)
                self.create_batch_files(site)
                self.estimate_positions(site)
                if(cleanup):
                    cleanup(site)

        os.chdir(self.main.home_dir)

#   ----------------------------------------           
    def cleanup(self, site):
        """
        Clean up temporary files used for processing.

        Args:
            site   = A station name (string) which belongs to
                     the list of the sites.                   
        """
        # remove the obsolete files in the processing directories
        if(self.processing_files_dirs != None):
            for fn in self.processing_files_dirs:
                os.system("rm -Rf " + fn + "/scratch")
                os.system("rm -Rf " + fn + "/*" + self.main.orbits_type[1])
                os.system("rm -Rf " + fn + "/track.log")
                os.system("rm -Rf " + fn + "/orbitfile.dat")

                ii = self.processing_files_dirs.index(fn)
                os.system("rm -Rf " + fn + "/" + self.processing_files[ii])
                os.system("rm -Rf " + fn + "/" + self.reference_station_files[ii])
# =====================================================================================
# EOF
