"""
This module provides the class GPSToRinexConverter. It allows to convert
raw Leica GPS500 data (mdb Format) into daily Rinex Files.
This class/module uses TEQC from Unavco.

"""

__author__    = "Patrick Riesen (riesen@vaw.baug.ethz.ch)"
__date__      = "2010-07-01"
__copyright__ = "Copyright (C) 2010 Patrick Riesen"
__license__   = "GNU GPL version 3 or any later version"

import sys
import os, getopt
import commands as cm
import numpy as n
import string as st
import time
from datetime import datetime as dt
from kingps.variables import execute

# Thanks:
# http://stackoverflow.com/questions/6451655/python-how-to-convert-datetime-dates-to-decimal-years
def toYearFraction(date):
    def sinceEpoch(date): # returns seconds since epoch
        return time.mktime(date.timetuple())
    s = sinceEpoch

    year = date.year
    startOfThisYear = dt(year=year, month=1, day=1)
    startOfNextYear = dt(year=year+1, month=1, day=1)

    yearElapsed = s(date) - s(startOfThisYear)
    yearDuration = s(startOfNextYear) - s(startOfThisYear)
    fraction = yearElapsed/yearDuration

    return date.year + fraction

# --------------------------------------------------------------------------------------------
class GPSToRinexConverter:
    """
    Class for Conversion of Raw Leica Data Files to Rinex Observation Files using TEQC.
    This class must be instantiated with a GPSProcessingVariables object as argument.
    
    """
#   ----------------------------------------           
    def __init__(self, gpsvariables):
        """ Initialize local variables for this class. """

        self.main = gpsvariables
        self.raw_files = []
        self.rinex_files = []
        self.file_decyear = []
        self.file_doy = []

#   ----------------------------------------           
    def get_input_files(self, site):
        """
        Create a list of the raw leica files which will be converted. This
        is done for all files of a specified site.

        Args:
            site = A station name (string) which belongs to
                   the list of the sites.
        """
        os.chdir(self.main.raw_input_dir + "/" + site)
        # get list of files
        self.raw_files = [] 
        for fn in os.listdir("."):
            name, ending = st.split(fn,".")
            if ending.startswith("o") or ending.startswith("O"):
                self.raw_files += [fn,]
        # filestring = cm.getoutput("ls -l -s -h *.[o,O]* | awk '{print $9}'")
        # self.raw_files = st.split(filestring, sep='\n')        
        os.chdir(self.main.home_dir)

#   ----------------------------------------           
    def convert(self, site, moving=False):
        """
        Convert Raw Leica Files to Rinex Files for one specified site.

        Args:
            site   = A station name (string) which belongs to
                     the list of the sites.
            moving = This is true for all sites, except for the
                     reference station. The Rinex Header will
                     include a tag for the data being
                     recorded as kinematic.                     
        """

        print "\n-----------------------------------------------------------"
        print "Converting Files to Rinex for site: ", site

        # get the files
        self.get_input_files(site)
        # change to output directory
        os.chdir(self.main.rinex_output_dir + "/" + site)
        
        # Create Rinex Files
        for fn in self.raw_files:
            # error file
            errfile = fn + ".err"

            # put command together
            c0 = " +err " + errfile + " -O.ag '" + self.main.institute \
                 + "' -O.o '" + self.main.observer + "' -O.mo " + site \
                 + " -lei mdb"
            if(moving):
                c0 = c0 + " -O.mov 1"
            obs_out = fn + ".o"
            c1 = " +obs " + obs_out + " "
            c2 = self.main.raw_input_dir + "/" + site + "/" + fn

            # run teqc on files
            command = self.main.teqc + c0 + c1 + c2
            self.main.thread(execute(command))

        # collect logging messages from teqc
        if(os.path.isfile(errfile)):
           os.system("cat *.err > log_teqcrx.out")
           os.system("rm -Rf *.err")

        os.chdir(self.main.home_dir)

#   ----------------------------------------           
    def rename(self, site):
        """
        Rename Rinex Files for specified site according to Rinxe convention.
        -> {site}{day}.{year}o for Observation Files.

        Args:
            site   = A station name (string) which belongs to
                     the list of the sites.                   
        """

        print "Renaming Files to Rinex for site: ", site

        # get rinex files in folder
        os.chdir(self.main.rinex_output_dir + "/" + site)
        files = []
        for fn in os.listdir("."):
            try:
                name, end1, end2  = st.split(fn,".")
            except Exception, e:
                print "Skipping file: ", fn
                pass
            else:
                if end1.startswith("o") or end1.startswith("O"):
                    if end2 == "o":
                        files += [ fn, ]

        # filestring = cm.getoutput("ls -l -s -h *.*.o | awk '{print $9}'")
        # files = st.split(filestring, sep='\n')

        # fetch decimal date from file header and fill in data
        self.file_doy = []
        self.file_decyear = []
        self.rinex_files = []

        for fn in files:
            # check if rinex file is not empty, else remove it
            isempty = cm.getoutput("ls -s -l " + fn + " | awk '{print $1}'")

            if(int(isempty) != 0):
                grep_date = "sed -n '/L1 & L2/{n;p;}' " + fn + \
                            " | awk '{print $1,$2,$3}'"

                datestr = cm.getoutput(grep_date)
                dateobj = dt.strptime(datestr,"%Y %m %d")                
                # get_doy = cm.getoutput("doy " + doy_in_file \
                #                        + " | grep 'DOY' | awk '{print $6}'")
                # get_doy = st.zfill(get_doy, 3)
                # self.file_doy.append(int(get_doy))
                doy = dateobj.strftime("%j")
                self.file_doy.append(doy)
                
                # get_year = cm.getoutput("doy " + cm.getoutput(grep_date) \
                #                         + " | grep 'Decimal' | awk '{print $3}'")
                self.file_decyear.append(float(toYearFraction(dateobj)))

                year_fid = str(dateobj.year)
                # year_fid=str(int(n.fix(float(get_year))))

                # rename file
                new_name = site + doy + "." + year_fid[-2:] + "o"
                self.rinex_files.append(new_name)
                os.rename(fn, new_name)
            else:
                print "\n*** Warning: Found empty file: ",fn,", removing it."
                os.system("rm -f " + fn)

        os.chdir(self.main.home_dir)

#   ----------------------------------------           
    def process(self, which="all"):
        """
        Convert and rename all or a single Raw Leica files
        to Rinex files for each site from the list
        of sites including the fix site.

        Args:
            which = A string which can be either 'all' or
                    a name of a site specified for processing.
        """
        # run through all sites
        if(which == "all"):
            self.convert(self.main.fix_name)
            self.rename(self.main.fix_name)

            if(self.main.site_names != None):
                for site in self.main.site_names:
                    self.convert(site, moving=True)
                    self.rename(site)
            os.chdir(self.main.home_dir)
        else:
            # try to process if it is a known site
            allsites =[self.main.fix_name]
            for site in self.main.site_names:
                allsites.append(site)
            try:
                idv = allsites.index(which)
            except ValueError:
                print "*** Error: This site was never specified (fix_name nor site_names)."
                raise ValueError
            else:
                if(idv > 0):
                    self.convert(which, moving=True)
                elif(idv == 0):
                    self.convert(which)                        
                self.rename(which)
                os.chdir(self.main.home_dir)

        os.chdir(self.main.home_dir)

class RinexUtils:
    """ This class includes some utilities for Rinex file processing."""
    def __init__(self, gpsvariables):
        self.main = gpsvariables

    def split(self, inputfile, output_prefix, window=("day",1)):
        """ Split a Rinex file into multiple files with a specified window length.

            Args:
                inputfile       = A string, path to Input file name.
                output_prefix   = A string, the path/prefix which should be used for the
                                  new files created.
                window          = A tuple (string, int) to indicate the splitting. E.g.
                                  ('day',1) [default] will produce files of 1 day length.
                                  Availabe are:
                                              ('day',1), ('hour',x),

            """
        if(window[0] == "day"):
            tc = " -tbin " + str(window[1]) + "d "
        elif(window[0] == "hour"):
            tc = " -tbin " + str(window[1]) + "h "
        elif(window[0] == "minute"):
            tc = " -tbin " + str(window[1]) + "m "
        elif(window[0] == "second"):
            tc = " -tbin " + str(window[1]) + "s "
        else:
            print "Sorry, There are no other possibilities than hour, day, minute or second."
            raise os.error()

        # Split files with teqc tbin command
        cmd = self.main.teqc + tc + output_prefix + " " + inputfile
        self.main.thread(execute(cmd))
# =================================================================================
# EOF
