"""
This modules includes a class called GPSProcessinVariables which handles
the GPs processing environment and its settings.

"""

__author__    = "Patrick Riesen (riesen@vaw.baug.ethz.ch)"
__date__      = "2010-07-01"
__copyright__ = "Copyright (C) 2010 Patrick Riesen"
__license__   = "GNU GPL version 3 or any later version"

import sys
import os, getopt
import commands as cm
import numpy as n
import string as st
from multiprocessing import Process, Pool

# --------------------------------------------------------------------------------------------
def execute(program, pinput=None, poutput=None):
    """
    Helper function to execute programs outside this python instance, e.g.,
    track and teqc and other shell commands.

    """
    if(pinput == None):
        command = program
    elif(poutput == None):        
        command = program + " " + pinput
    else:
        command = program + " " + pinput + " > " + poutput

    os.system(command)

#---------------------------------------------------------------------------------------------        
class GPSProcessingVariables:
    def __init__(self):
        """
        Initialize Environment. Set processing variables and
        start a thread which will execute batch commands

        Instantiating this class takes no input arguments.

        """

        # Determine maximum cpus
        #cps = max(cm.getoutput("sed -n '/processor/p' /proc/cpuinfo | awk '{printf /%d/ $3}'"))
        #cps = int(cps)+1
        cps = int(1)
        self.pool = Pool(cps)
        self.cpulist = n.arange(0,cps)
        self.home_dir = os.getcwd()

        # Some Variables:
        # ---------------------------------------- 

        # GPS DATA
        self.observer = "Patrick Riesen" 
        self.institute = "VAW ETH Zuerich"
        self.fix_xyz = None
        self.fix_name = "fix1"
        self.N_sites = 0
        self.site_names = None

        # Programs
        self.source_file = "conf.dat"
        self.teqc = "nice -19 teqc"
        self.track = "nice -19 track"
        self.orbits = "nice -19 sh_get_orbits"
        self.orbits_archive = "cddis"
        self.orbits_type = ("igs","sp3")
        self.shell = "/usr/bin/tcsh"

        # IO
        self.prefix_dir = None
        self.raw_input_dir = None
        self.rinex_output_dir = None
        self.processing_output_dir = None
        self.orbit_files_dir = None

        # Track
        self.site_stats = (20, 0.01)
        self.output = "GEOD"
        self.mode = "short"
        self.interval = 120       
        self.smooth = True
        self.out_sig_limit = 0.05
        self.cut_off = 15
        self.ambiguity = False

#   ----------------------------------------
    def thread(self, execute_func):
        """
        Execute a function in the thread.

        Args:
            execute_func = A global function

        """
        self.pool.map(execute_func, self.cpulist.tolist())

#   ----------------------------------------
    def finish(self):
        """ Close Sub-Thread """
        self.pool.terminate()

#   ----------------------------------------
    def init_sites(self, site_names):
        """
        Set the list of site names.

        Args:
            site_names = A python list of site_names (strings)

        """
        self.site_names = site_names
        self.N_sites = int(len(site_names))

#   ----------------------------------------
    def init_workspace(self, prefix=None):
        """
        Create the processing folder structure under the specified directory.

        Args:
            prefix = A directory (string)

        """

        if(prefix != None):
            # Initialize folder structure
            if(os.path.isdir(prefix) == False):
                os.mkdir(prefix)
            else:
                print "\n*** Warning: Workspace Location exists already."
                print "    Leave it as is."

        # check if path has been set already and set default paths
        if(self.prefix_dir != prefix):
            self.prefix_dir = prefix

        self.raw_input_dir = prefix + "/raw"
        if(os.path.isdir(self.raw_input_dir) == False):
            os.mkdir(self.raw_input_dir)

        self.rinex_output_dir = prefix + "/rinex"
        if(os.path.isdir(self.rinex_output_dir) == False):
            os.mkdir(self.rinex_output_dir)

        self.processing_output_dir = prefix + "/processed"
        if(os.path.isdir(self.processing_output_dir) == False):
            os.mkdir(self.processing_output_dir)

        self.orbit_files_dir = prefix + "/orbits"
        if(os.path.isdir(self.orbit_files_dir) == False):
            os.mkdir(self.orbit_files_dir)

        if(self.site_names != None):
            self.N_sites = len(self.site_names)
            for path in (self.raw_input_dir, self.rinex_output_dir, self.processing_output_dir):
                for site in self.site_names:
                    site_path = path + "/" + site
                    exists = os.path.isdir(site_path)
                    if(exists == False):
                        os.mkdir(site_path)
        else:
            print "\n*** Error: Site Names have not been defined yet."
            raise os.error()

        # init fix
        for path in (self.raw_input_dir, self.rinex_output_dir, self.processing_output_dir):
            site_path = path + "/" + self.fix_name
            exists = os.path.isdir(site_path)
            if(exists == False):
                os.mkdir(site_path)

        #rm one path not needed
        os.system("rmdir " + self.processing_output_dir + "/" + self.fix_name)
        
        print "\n-----------------------------------------------------------"
        print "Created Workspace:\n"
        os.system("ls -1 -h " + prefix + "/*")
        print "\n-----------------------------------------------------------"
        print "\n*** Attention: Don't forget to fill in the raw data in the correct paths before you"
        print "    call any further routines, e.g. Rinex file creation or even Processing."
        print "    Or else, you may call search_rawdata() now."
        # change back to top directory
        os.chdir(self.home_dir)  
        
#   ----------------------------------------           
    def search_rawdata(self, search_dir=None):
        """
        Search the specified folder for raw data files recursively and copy the data
        found to prefix_dir/raw/... for the sites given in site_names.

        Args:
            search_dir = A full path to a directory to search recursively (string)

        """
        if(search_dir == None):
            print "\n*** Error: If you want to search a folder, you should specify the folder."
            raise os.error()
        
        # check it site_names was defined
        if(self.site_names == None):
            print "\n*** Error: You did not specify the site names, don't know how to sort"
            print "      data potentially found in the folders to search."
            raise os.error()
        else:
            # search for data in subfolders name foo/site_name[i] and fix
            allsites = [self.fix_name]
            for sa in self.site_names:
                allsites.append(sa)

            os.chdir(self.home_dir)
            for site in allsites:
                sdir = search_dir + "/" + site                
                os.chdir(sdir)            
                filestring = cm.getoutput("find . -name '*.[O,o]*'")
                files = st.split(filestring, sep='\n')

                # if files found, copy them to input folder
                if(len(files) > 0):
                    print "\nFound ",len(files),"files for site",site
                    if(self.raw_input_dir == None):
                        print "\n*** Error: Don't know where to put the files. Did you"
                        print "    call init_workspace() before this routine? Or did you"
                        print "    forget to specify the variable raw_input_dir?"
                    else:
                        putdir = self.raw_input_dir + "/" + site
                        for fn in files:
                            cmd = "cp " + fn + " " + putdir
                            print cmd
                            self.thread(execute(cmd))
                else:
                    print "\n*** Warning: No data found for site", site,"!"
                    
        os.chdir(self.home_dir)
        
#   ----------------------------------------           
    def erase_data(self,
                   raw=False, rinex=False,
                   orbits=False, processed=False):
        """
        Erase the files within ceratin folders of the workspace.

        Args:
            raw       = True if data under prefix/raw should be removed.
            rinex     = True if data under prefix/rinex should be removed.
            orbits    = True if all downloaded orbit files should be removed.
            processed = True if all data under prefix/processed should be removed.

        Per default, all argumens are set to False, you must explicitly
        specify the boolean values to remove the desired content.

        """
        paths = (self.raw_input_dir, self.rinex_output_dir,
                 self.processing_output_dir)
        values = (raw, rinex, processed)
        for path, doit in zip(paths, values):
            if(doit):
                if(path == None):
                    if(self.prefix_dir == None):
                        print "\n*** Error: Don't know which paths to erase,"
                        print "    no working directory (prefix_dir) specified."
                        raise os.error()
                    else:
                        if(doit == raw):
                            path = self.prefix_dir + "/raw"
                        elif(doit == rinex):
                            path = self.prefix_dir + "/rinex"
                        elif(doit == processed):
                            path = self.prefix_dir + "/processed"

                allsites = [self.fix_name]
                for site in self.site_names:
                    allsites.append(site)
                for site in allsites:
                    site_path = path + "/" + site

                    erase_data = "rm -Rf " + site_path
                    os.system(erase_data)
        if(orbits):
            os.system("rm -Rf " + self.orbit_files_dir + "/*")

        os.chdir(self.home_dir)
# ==============================================================================
# EOF
